"use strict";
this.name	     = "Synchronised Torus Controller";
this.version     = "0.91";    // beta release 2015-12-11
this.description = "equipment script for torus drive synchronisation";
this.author	     = "Fritz G.";   
this.copyright	 = "© 2015 Fritz G.";
this.licence	 = "CC BY-NC-SA 4.0";

//------------------------------------------------------------------------------------------------------------
// Credits to Capt. Murphy for his idea to implement torus synchronisation in his Escort Contracts OXP.
//------------------------------------------------------------------------------------------------------------
// This OXP is based on an idea I had when escorting slow freighters. Because NPX ships have no torus drive,
// this can take very long, if you want to follow them all the way from witchpoint to station.
// Later I discovered Escort Contracts OXP, using "torus synchronisation" to speed up the game.
// This of course works only for the ship you have agreed to escort, but using the same basic technique
// (simulating torus drive by using a high velocity) I was able to make this available for all ships 
// that are underway in a straight line to a distant destination. That could be traders, bounty hunters,
// vipers or even pirates or assassins on their way to the station, the witchpoint or the sun. 
// In theory this should work with all ships regardless of role, including OXP ships. Only Thargoids are 
// excluded but they usually attack and dont't fly to a distant destination. 
// 
// Because the player shouldn't be able to change the destination of a NPC ship, the direction of torus 
// driving (and its speed) is defined by the NPC. The torus drive stops if the destination is reached or  
// if either the player or the mother is mass locked. 
//------------------------------------------------------------------------------------------------------------

// configuration (values taken from Escort Contracts v1.7.1)
this.$st_minDist     	= 500;	// minimum distance to NPC ship in game meters
this.$st_maxDist     	= 2500;	// maximum distance to NPC ship in game meters
this.$st_headingAlign  	= 0.98;	// precision needed for heading alignment (1 is exact (impossible!), 0.9 is too easy and looks odd) 
//------------------------------------------------------------------------------------------------------------
this.$st_torusEngaged = false;  // this can be used read only (!) by other OXPs (similar to player.ship.torusEngaged)    
//------------------------------------------------------------------------------------------------------------
this.$st_timerRunning = false;  
this.$st_npc = null;           // will hold the ship object of the npc 
this.$st_scanClass = ""; 	   // for saving scanner properties of targeted npc ship	 

// key n pressed for starting or stopping torus synchronisation
this.activated = function ()
{
    if (!this.$st_timerRunning) 
	{
		// initialize timer 
		// The Escort Contracts AI checks every second, so that should be enough here too.
		this.$st_timer =  new Timer(this, this.$st_torusDriving, -1, 1);  
		this.$st_timerRunning = true;
	}	

	if (this.$st_torusEngaged) 
	{
		// stop synchronised torus 
        this.$st_stopTorus(1);
		player.consoleMessage("Synchronised torus stopped.");
	}
	else
	{	
		// try to start synchronised torus 
		// first check the conditions (NPC targeted, correct distance and heading, NPC destination, mass locking)
		if (this.$st_checkConditions(1))
		{
			player.consoleMessage("Synchronised torus drive engaged.");
			// Because the core game doesn't feature torus drives for npc ships, we have to simulate 
			// it by a high velocity. The factor 32 is the same as used for the standard torus drive.
			// Synchronised torus speed is defined by the npc ship because it is the leader.
			// For exact synchronisation of velocities the timer interval is too long, so use frame callback.	
			this.$st_frameCallback = addFrameCallback(this.$st_matchVelocity.bind(this));   
			this.$st_npc.velocity = this.$st_npc.heading.multiply(this.$st_npc.maxSpeed * 32); 
			// velocity will be synchronised in this.$st_matchVelocity(), but if we dont't do it here, the npc  
			// will start up to 1/60s (frame rate) earlier, and this is visible because of the very high speed.
			player.ship.velocity = this.$st_npc.velocity;
			this.$st_torusEngaged = true;
			this.$st_timer.start();      
		}
		else
		{
			// Initialisation failed, reset the npc scanner class.
			this.$st_resetScanClass();
		}	
	}	
};

// key b pressed (currently not used)
// this.mode = function () {};

// Check if conditions allow starting or continuing torus synchronisation. 
// Function called on activation (n key) (1) and in timer function (2).
this.$st_checkConditions = function(calledBy)
{
	var ps = player.ship;
    if (calledBy === 1)
    {
		// Checks only needed when starting (conditions shouldn't change after starting).
		// First we have to check if something is targeted,
		if (!ps.target) 
		{
			player.consoleMessage("No target selected.");
			return false;
		}		
		this.$st_npc = ps.target;
		// Exclude unsuitable "ships"
		// An Escort Contracts mother is unsuitable because the results could be unpredictable...
		// Usually the mother starts synchronised torus driving anyway if it is possible.
		if (!this.$st_npc.isPiloted || this.$st_npc.isThargoid || this.$st_npc.isStation || this.$st_npc.primaryRole === "ec_mother" || this.$st_npc.primaryRole === "escape-capsule") 
		{
			player.consoleMessage("Unsuitable target.");
			return false;
		}		
		// Does the target move?
		if (this.$st_npc.speed === 0)			
        {
			player.consoleMessage("Target is stationary.");
			return false;
		}	
		// Check distance between ships 
		if(this.$st_npc.position.distanceTo(ps.position) > this.$st_maxDist)
		{
			player.consoleMessage("Distance to target must not exceed  " + this.$st_maxDist + " m.");
			return false;
		}			
		else if(this.$st_npc.position.distanceTo(ps.position) < this.$st_minDist)
		{
			player.consoleMessage("Distance to target must be at least " + this.$st_minDist + " m.");
			return false;
		}			
		// To prevent the player being mass-locked we have to give the npc the scan class "rock". 
		// $st_setScanClass() will also turn the scanner lollipop to flashing yellow and magenta (like in Escort Contracts).
		// Side effect: The legal status will not be shown during torus driving when using the Scanner Targeting Enhancement.
		// Escorts, if present, must be "rocked" too, but they don't need a different colour.
		this.$st_setScanClass(); 
    }

    // The following conditions are also checked continuously during torus driving

	// Torus synchronisation makes only sense over larger distances, so the distance to  
	// destination of npc ship must be outside scanner range.
	if (this.$st_npc.position.distanceTo(this.$st_npc.destination) < ps.scannerRange)			
	{
		if (calledBy === 1) 
			player.consoleMessage("Next waypoint too close.");
		else
			player.consoleMessage("Waypoint reached.");
		return false;
	}	
	// The player can change heading (but not velocity) during torus driving, 
	// so this has to be checked continuously. 
	if(this.$st_npc.heading.dot(ps.heading) < this.$st_headingAlign)
	{
		player.consoleMessage("Heading mismatch.");
		return false;
	}			
	// Check for player mass lock (check returns true when done immediately after changing the npc ship to rock).
	if (calledBy != 1 && player.alertMassLocked) 
	{
		player.consoleMessage("Mass-locked.");
		return false;
	}	
	// Check for npc mass lock. 
	var targets = system.filteredEntities(this, st_checkForMassLock, this.$st_npc, this.$st_npc.scannerRange);
	if (targets.length > 0) 
	{
		player.consoleMessage("Target mass-locked.");
		return false;
	}	
	// If player engages normal torus drive or injectors, synchronisation will be stopped.
	// Player will be mass locked by npc ship shortly after this. 
	// Note: Torus driving does _not_ change speed, but using injectors does!
	if (ps.torusEngaged || ps.speed > ps.maxSpeed) 
	{
		return false;
	}	
	// Everything is (still) ok  
	return true;
};	

// Used to check for entities other than the player mass-locking the npc.
this.st_checkForMassLock = function(entity) 
{
    // This should exclude the escorts too, but it doesn't yet.
	// Distance to planet or sun isn't checked; this is a little bug in Escort Contracts too. 
	
	return ((entity.scanClass === "CLASS_NEUTRAL" || entity.scanClass === "CLASS_MILITARY" || entity.scanClass === "CLASS_POLICE" || entity.scanClass === "CLASS_THARGOID" || entity.scanClass === "CLASS_STATION") && !entity.isPlayer);
};

// stop torus (called by pressing n key (1) or by timer function (2) 
this.$st_stopTorus = function(calledBy)
{
	this.$st_timer.stop();
	this.$st_torusEngaged = false;      
    // Stop velocity synchronisation.
	this.$st_removeFrameCallback(); 	
	// Bring ships back to normal speed.
	this.$st_npc.velocity = this.$st_npc.thrustVector; 
	player.ship.velocity = player.ship.thrustVector; 
    // Reset scan class and lollipop colours.	
	this.$st_resetScanClass(); 
};	

// Set scan class to prevent mass locking.
this.$st_setScanClass = function()
{
	if (this.$st_npc)
	{	
		this.$st_scanClass = this.$st_npc.scanClass;  // save for resetting		 

		this.$st_npc.scanClass = "CLASS_ROCK"; 		
		this.$st_npc.scannerDisplayColor1 =	"yellowColor";	
		this.$st_npc.scannerDisplayColor2 =	"magentaColor";		
	}	
};

// Reset scan class after stopping torus driving or after a failed starting attempt.
this.$st_resetScanClass = function()
{
	// only reset if it has been changed already
	if (this.$st_npc && this.$st_scanClass != "")
	{	
		this.$st_npc.scanClass = this.$st_scanClass; 		
		this.$st_npc.scannerDisplayColor1 =	null; 		
		this.$st_npc.scannerDisplayColor2 =	null;		
		this.$st_scanClass = "";
	}	
};

// Timer function, checks if conditions are still ok.
this.$st_torusDriving = function()
{
	if (!this.$st_checkConditions(2))
	{
		this.$st_stopTorus(2);
	}
	else
	{
		// Velocity reduces slowly, so we have to keep it up.
		this.$st_npc.velocity = this.$st_npc.heading.multiply(this.$st_npc.maxSpeed * 32); 
		player.ship.velocity = this.$st_npc.velocity;
	}		
};
	
// Frame callback function. 
this.$st_matchVelocity = function() 
{
	player.ship.velocity = this.$st_npc.velocity;
};	

// Remove frame callback function. 
this.$st_removeFrameCallback = function() 
{
	if (this.$st_frameCallback)
	{
		removeFrameCallback(this.$st_frameCallback);
		delete this.$st_frameCallback;
	}
};
